@@ -8,12 +8,13 @@ our @EXPORT = qw(ServiceNow);
 use SOAP::Lite;
 use LWP::UserAgent;
 use HTTP::Cookies;
+use File::Basename;
 use MIME::Base64;
 use XML::Simple;
 use Time::HiRes;
 use Carp;
 
-our $VERSION = '0.10';
+our $VERSION = '0.11';
 
 =head1 NAME
 
@@ -261,7 +262,8 @@ sub new {
         dv => 0,
         fetch => 250,
         query => 0,
-        client => $client
+        client => $client,
+        private => {}
     };
     bless $session, $pkg;
     $session->set(@options);
@@ -271,6 +273,21 @@ sub new {
     return $session;
 }
 
+#
+# private method to return a private table for use by this module
+#
+sub private {
+    my ($session, $tablename) = @_;
+    my $table = $session->{private}{$tablename};
+    if (!$table) {
+        $table = $session->table($tablename);
+        $session->{private}{$tablename} = $table;
+    }
+    # make sure DV is always false for private tables
+    $table->setDV(0);
+    return $table;
+}
+
 our $startTime;
 
 sub traceBefore {
@@ -310,7 +327,7 @@ Use of this method requires the C<soap_script> role and activation of the
 B<Syntax>
 
     my %outputs = $sn->call($name, %inputs);
-    my $outputs = $sn->call($name, $inputs);
+    my $outputs = $sn->call($name, $inputs); # $inputs must be a hash reference
 
 B<Example>
 
@@ -417,9 +434,8 @@ sub saveSession {
 =head2 set
 
 This method permits assignment of various default values and processing options.
-This method is generally not required
-since the same options can be specified 
-following the password parameter of L</ServiceNow>.
+The same options can also be specified
+following the password argument of L</ServiceNow>.
 
 The following options may be specified.
 
@@ -445,7 +461,7 @@ by a parameter to L</fetch> or L</fetchAll>.
 C<query> - Specify the number of keys to be retrieved by
 L</query> in each call to L</getKeys>.
 The default is 0,
-which causes all key to be retrieved 
+which causes all keys to be retrieved 
 in a single call.
 
 =item *
@@ -496,7 +512,8 @@ sub setTrace {
 
 =head2 soap
 
-This method returns a reference to the underlying SOAP::Lite object.
+This method returns a reference to the underlying 
+L<SOAP::Lite> object.
 
 =cut
 
@@ -536,7 +553,7 @@ L<Direct Web Services API|http://wiki.servicenow.com/index.php?title=SOAP_Direct
 To obtain a Table object,
 use the L</table> method describe above.
 
-These Table object methods provide access to the Direct Web Services API:
+The following Table object methods provide access to Direct Web Services API methods of the same name:
 L</deleteRecord>,
 L</get>,
 L</getKeys>,
@@ -589,7 +606,7 @@ sub callMethod {
     my $client = $session->{client};
     my $baseurl = $session->{url};
     my $endpoint = "$baseurl/$tablename.do?SOAP";
-    if (($methodname eq 'get' || $methodname eq 'getRecords') && $self->{dv}) {
+    if ($methodname =~ /^(get|getRecords|aggregate)$/ && $self->{dv}) {
         my $dv = $self->{dv};
         $endpoint = "$baseurl/$tablename.do?displayvalue=" . $dv . "&SOAP";
     }
@@ -618,6 +635,42 @@ sub traceAfter {
     $session->traceAfter($trace, $message, $content);
 }
 
+=head2 addComment
+
+This method adds a comment to a task.
+
+See also: L</getComments>, L</getJournalEntries>
+
+B<Syntax>
+
+    $table->addComment($sys_id, $comment);
+    
+=cut
+
+sub addComment {
+    my ($self, $sysid, $comment) = @_;
+    $self->update($sysid, 'comments' => $comment);
+    return $self;
+}
+
+=head2 addWorkNote
+
+This method adds a work note to a task.
+
+See also: L</getWorkNotes>, L</getJournalEntries>
+
+B<Syntax>
+
+    $table->addWorkNote($sys_id, $work_note);
+    
+=cut
+
+sub addWorkNote {
+    my ($self, $sysid, $note) = @_;
+    $self->update($sysid, 'work_notes' => $note);
+    return $self;
+}
+
 =head2 asQuery
 
 This method creates a new L<Query|/ServiceNow::SOAP::Query> object
@@ -669,57 +722,68 @@ This method implements the
 L<attachment creator API|http://wiki.servicenow.com/index.php?title=AttachmentCreator_SOAP_Web_Service>.
 The method requires C<soap_ecc> role.
 
-You will need to specify a MIME type.
-If no type is specified,
-this function will use a default type of "text/plain".
-For a list of MIME types, refer to
-L<http://en.wikipedia.org/wiki/Internet_media_type>.
-
 When you attach a file to a ServiceNow record,
-you can also specify an attachment name
+you can optionally specify an attachment name
 which is different from the actual file name.
 If no attachment name is specified,
 this function will
-assume that the attachment name is the same as the file name.
+assume that the attachment name is the 
+the file name stripped of any directory.
+
+You can also optionally specify a MIME type.
+If no type is specified,
+the function will use L<MIME::Types>
+to infer the type.
 
 This function will die if the file cannot be read.
 
 B<Syntax>
 
-    $table->attachFile($sys_id, $filename, $mime_type, $attachment_name);
+    $table->attachFile($sys_id, $filepath);
+    $table->attachFile($sys_id, $filepath, $attachment_name, $mime_type);
 
 B<Example>
 
-    $incident->attachFile($sys_id, "screenshot.png", "image/png");
+    $incident_tbl->attachFile($sys_id, "screenshot.png");
     
 =cut
 
 sub attachFile {
     my $self = shift;
-    my ($sysid, $filename, $mimetype, $attachname) = @_;
-    Carp::croak "attachFile: No such record" unless $self->get($sysid);
-    $mimetype = "text/plain" unless $mimetype;
-    $attachname = $filename unless $attachname;
-    Carp::croak "Unable to read file \"$filename\"" unless -r $filename;
     my $session = $self->{session};
-    my $ecc_queue = $self->{ecc_queue} ||
-        ($self->{ecc_queue} = $session->table('ecc_queue'));
-    open my $fh, '<', $filename 
-        or die "Unable to open \"$filename\"\n$!";
+    my $tablename = $self->{name};
+    my ($sysid, $filepath, $attachname, $mimetype) = @_;
+    Carp::croak "attachFile: No such record" unless $self->get($sysid);
+    my ($parsename, $parsepath, $suffix ) = 
+        File::Basename::fileparse($filepath, "\.[^.]*");
+    if (!$attachname) {
+        $attachname = $parsename . $suffix;
+    }
+    if (!$mimetype) {
+        require MIME::Types;
+        my $typer = MIME::Types->new;
+        $mimetype = $typer->mimeTypeOf($suffix);
+    }
+    Carp::croak "Unable to read file \"$filepath\"" unless -r $filepath;
+    my $ecc_queue = $session->private('ecc_queue');
+    open my $fh, '<', $filepath 
+        or die "Unable to open \"$filepath\"\n$!";
+    binmode $fh;
     my ($buf, $base64);
     # encode in multiples of 57 bytes to ensure no padding in the middle
     while (read $fh, $buf, 60*57) {
         $base64 .= MIME::Base64::encode_base64($buf);
     }
     close $fh;
-    my $tablename = $self->{name};
-    my $sysid = $ecc_queue->insert(
+    my $len = length($base64);
+    print "attachment name=$attachname:$mimetype len=$len\n";
+    my $ecc_id = $ecc_queue->insert(
         topic => 'AttachmentCreator',
         name => "$attachname:$mimetype",
         source => "$tablename:$sysid",
         payload => $base64
     );
-    die "attachFile failed" unless $sysid;
+    die "attachFile failed" unless $ecc_id;
 }
 
 =head2 columns
@@ -750,11 +814,13 @@ or the number of records that match a set of parameters.
 This method requres installation of the
 L<Aggregate Web Service plugin|http://wiki.servicenow.com/index.php?title=SOAP_Direct_Web_Service_API#aggregate>.
 
+See also: L</countBy>
+
 B<Syntax >
 
     $count = $table->count();
+    $count = $table->count($encodedQuery);
     $count = $table->count(%parameters);
-    $count = $table->count($encodedquery);
 
 B<Examples>
 
@@ -773,11 +839,60 @@ sub count {
     my @params = (COUNT => 'sys_id', @_);
     $self->traceBefore('aggregate count');
     my $som = $self->callMethod('aggregate' => _params @params);
-    my $count = $som->body->{aggregateResponse}->{aggregateResult}->{COUNT};
+    my $count = $som->body->{aggregateResponse}{aggregateResult}{COUNT};
     $self->traceAfter("count=$count");
     return $count;
 }
 
+=head2 countBy
+
+This method counts the number of records in a table
+using a GROUP_BY parameter,
+which must be the name of a column in the table.
+As with L</count>,
+an encoded query or set of filter parameters can be specified.
+The method returns a list of name/value pairs.
+
+This method requres installation of the
+L<Aggregate Web Service plugin|http://wiki.servicenow.com/index.php?title=SOAP_Direct_Web_Service_API#aggregate>.
+
+See also: L</count>
+
+B<Syntax>
+
+    %counts = $table->countBy($groupByColumn);
+    %counts = $table->countBy($groupByColumn, $encodedQuery);
+    %counts = $table->countBy(%groupByColumn, %parameters);
+
+B<Example>
+
+    my %counts = $sn->table("incident")->countBy("category");
+    foreach my $key (keys %counts) {
+        my $count = $counts{$key};
+        print "category: $key, count: $count\n";
+    }
+
+=cut
+
+sub countBy {
+    my $self = shift;
+    my $groupby = shift;
+    Carp::croak "countBy missing groupby field" unless $groupby;
+    my @params = (COUNT => 'sys_id', GROUP_BY => $groupby, @_);
+    $self->traceBefore('aggregate count');
+    my $som = $self->callMethod('aggregate' => _params @params);
+    my @aggResults = @{$som->body->{aggregateResponse}{aggregateResult}};
+    my $rows = @aggResults;
+    $self->traceAfter("rows=$rows");
+    my %results;
+    foreach my $agg (@aggResults) {
+        my $key = $agg->{$groupby};
+        my $count = $agg->{COUNT};
+        $results{$key} = $count;
+    }
+    return %results;
+}
+
 =head2 deleteRecord
 
 This method deletes a record using the sys_id.
@@ -817,6 +932,8 @@ is to that you can pass the result back in as an
 C<__exclude_columns>
 L<extended query parameter|http://wiki.servicenow.com/index.php?title=Direct_Web_Services#Extended_Query_Parameters>.
 
+See also: L</exclude>, L</include>
+
 B<Syntax>
 
     $names = $table->except(@list_of_columns);
@@ -829,11 +946,11 @@ from the incident table.
 In other words, 
 the result excludes all columns B<except> sys_id, number and description.
 
-    my $incident = $sn->table("incident");
-    my @recs = $incident->getRecords(
+    my $incident_tbl = $sn->table("incident");
+    my @recs = $incident_tbl->getRecords(
         __encoded_query => "sys_created_on>=$datetime",
         __exclude_columns => 
-            $incident->except("sys_id,number,description"),
+            $incident_tbl->except("sys_id,number,description"),
         __limit => 250);
  
 =cut
@@ -869,7 +986,7 @@ B<Example>
 sub get {
     my $self = shift;
     my $tablename = $self->{name};
-    # Carp::croak("get $tablename: no parameter") unless @_; 
+    Carp::croak("get $tablename: missing argument") unless @_; 
     if (_isGUID($_[0])) { unshift @_, 'sys_id' };
     $self->traceBefore('get');
     my $som = $self->callMethod('get' =>  _params @_);
@@ -878,6 +995,59 @@ sub get {
     return $result;
 }
 
+=head2 getComments
+
+This method retrieves all comments for a task.
+It returns an array of records, each of which contains a C<value> field.
+
+See also: L</addComment>, L</getJournalEntries>
+
+B<Example>
+
+    my @comments = $incident_tbl->getComments($sys_id);
+    foreach my $comment (@comments) {
+        print $comment->{value}, "\n";
+    }
+    
+=cut
+
+sub getComments {
+    my ($self, $sysid) = @_;
+    return $self->getJournalEntries($sysid, 'comments');
+}
+
+=head2 getJournalEntries
+
+This method retrieves journal fields
+(I<i.e.> Work Notes, Comments) for a ticket.
+
+See also: L</getComments>, L</getWorkNotes>
+
+B<Syntax>
+
+    @notes = $table->getJournalEntries($sys_id, $fieldname);
+    
+B<Example>
+
+    my @notes = $incident_tbl->getJournalEntries($sys_id, 'work_notes');
+    foreach my $note (@notes) {
+        print $note->{value}, "\n";
+    }
+    
+=cut
+
+sub getJournalEntries {
+    my ($self, $sysid, $fieldname) = @_;
+    my $session = $self->{session};
+    my $sys_journal_field = $session->private('sys_journal_field');
+    $fieldname = 'work_notes' unless $fieldname;
+    my @params = (element_id => $sysid);
+    push @params, 'element' => $fieldname if $fieldname;
+    push @params, '__limit' => 1000;
+    my @recs = $sys_journal_field->getRecords(@params);
+    return @recs;
+}
+
 =head2 getKeys
 
 This method returns a list of keys.
@@ -1011,247 +1181,14 @@ sub getRecords {
     return @records;
 }
 
-=head2 insert
+=head2 getName
 
-This method inserts a record.
-In a scalar context, this method returns a sys_id only.
-In a list context, this method returns a list of name/value pairs.
-
-For information on available parameters
-refer to the ServiceNow documentation on the 
-L<"insert" Direct SOAP API method|http://wiki.servicenow.com/index.php?title=SOAP_Direct_Web_Service_API#insert>.
-
-B<Syntax>
-
-    $sys_id = $table->insert(%values);
-    %result = $table->insert(%values);
-    
-B<Examples>
-
-    my $incident = $sn->table("incident");
-
-In a scalar context, the function returns a sys_id.
-
-    my $sys_id = $incident->insert(
-        short_description => $short_description,
-        assignment_group => "Network Support",
-        impact => 3);
-    print "sys_id=", $sys_id, "\n";
-
-In a list context, the function returns a list of name/value pairs.
-
-    my %result = $incident->insert(
-        short_description => $short_description,
-        assignment_group => "Network Support",
-        impact => 3);
-    print "number=", $result{number}, "\n";
-    print "sys_id=", $result{sys_id}, "\n";
-
-You can also call it this way:
-
-    my %rec;
-    $rec{short_description} = $short_description;
-    $rec{assignment_group} = "Network Support";
-    $rec{impact} = 3;
-    my $sys_id = $incident->insert(%rec);
-
-Note that for reference fields (e.g. assignment_group, assigned_to)
-you may pass in either a sys_id or a display value.
+This method returns the name of this table from L</table>.
 
 =cut
 
-sub insert {
-    my $self = shift;
-    $self->traceBefore('insert');
-    my $som = $self->callMethod('insert', _params @_);
-    my $response = $som->body->{insertResponse};
-    my $sysid = $response->{sys_id};
-    $self->traceAfter();
-    return wantarray ? %{$response} : $sysid;
-}
-
-=head2 insertMultiple
-
-This method inserts multiple records.
-The input is an array of hash references.
-It returns an array of hash references.
-
-This method requires installation of the 
-L<ServiceNow Insert Multiple Web Services Plugin|http://wiki.servicenow.com/index.php?title=Web_Service_Import_Sets#Inserting_Multiple_Records>.
-
-For additional information 
-refer to the ServiceNow documentation on the 
-L<"insertMultiple" Direct SOAP API method|http://wiki.servicenow.com/index.php?title=SOAP_Direct_Web_Service_API#insertMultiple>.
-
-B<Syntax>
-
-    @results = $table->insertMultiple(@records);
-
-B<Example>
-
-    my $incident = $sn->table("incident");
-    my @allrecs;
-    foreach my $descr (@descriptions) {
-        my $newrec = {
-            short_description => $desc,
-            assignment_group => $groupName,
-            description => $desc,
-            impact => 3};
-        push @allrecs, $newrec;
-    }
-    my @results = $incident->insertMultiple(@allrecs);
-    foreach my $result (@results) {
-        print "inserted ", $result->{number}, "\n";
-    }
-
-=cut
-
-sub insertMultiple {
-    my $self = shift;
-    $self->traceBefore('insertMultiple');
-    my @params;
-    foreach my $rec (@_) {
-        Carp::croak "Expected array of hashs" unless ref $rec eq 'HASH';
-        push @params, SOAP::Data->name('record' => $rec);
-    }
-    my $som = $self->callMethod('insertMultiple', @params);
-    my $response = $som->body->{insertMultipleResponse};
-    my $insertResponse = $response->{insertResponse};
-    my @status = @{$insertResponse};
-    my $count = scalar(@status);
-    $self->traceAfter("$count records inserted");
-    return @status;
-}
-
-=head2 update
-
-This method updates a record.
-For information on available parameters
-refer to the ServiceNow documentation on the 
-L<"update" Direct SOAP API method|http://wiki.servicenow.com/index.php?title=SOAP_Direct_Web_Service_API#update>.
-
-B<Syntax>
-
-    $table->update(%parameters);
-    $table->update($sys_id, %parameters);
-
-Note: If the first syntax is used, 
-then the C<sys_id> must be included among the parameters.
-If the second syntax is used, then the C<sys_id>
-must be the first parameter.
-
-For reference fields (e.g. assignment_group, assigned_to)
-you may pass in either a sys_id or a display value.
-
-B<Examples>
-
-The following three examples are equivalent.
-
-    $incident->update(
-        sys_id => $sys_id, 
-        assigned_to => "5137153cc611227c000bbd1bd8cd2005",
-        incient_state => 2);
-        
-    $incident->update(
-        sys_id => $sys_id, 
-        assigned_to => "Fred Luddy", 
-        incident_state => 2);
-    
-    $incident->update($sys_id, 
-        assigned_to => "Fred Luddy", 
-        incident_state => 2);
-
-=cut
-
-sub update {
-    my $self = shift;
-    if (_isGUID($_[0])) { unshift @_, 'sys_id' };
-    my %values = @_;
-    my $sysid = $values{sys_id};
-    $self->traceBefore("update");
-    my $som = $self->callMethod('update', _params @_);
-    $self->traceAfter();
-    return $self;
-}
-
-=head2 query
-
-This method creates a new L<Query|/ServiceNow::SOAP::Query> object
-by calling L</getKeys>.
-You can then use L</fetch> or L</fetchAll>
-to retrieve the actual record.
-
-This method normally fetches all keys using a single 
-call to L</getKeys>.
-However, if the number of keys is too large you may
-encounter timeout errors. In this situation the recommendation
-is to fetch the keys in chunks by specifying 
-a C<query> option in the L</ServiceNow> function.
-The suggested initial setting is C<query =E<gt> 100000>.
-If timeouts are still a problem the reduce the value further.
-
-B<Syntax>
-
-    $query = $table->query(%parameters);
-    $query = $table->query($encodedquery);
-    
-B<Example>
-
-The following example builds a list of all Incidents 
-created between 1/1/2014 and 2/1/2014 sorted by creation date/time.
-The L</fetchAll> method fetches chunks of records 
-until all records have been retrieved.
-
-    my $filter = "sys_created_on>=2014-01-01^sys_created_on<2014-02-01";
-    my $qry = $sn->table("incident")->query(
-        __encoded_query => $filter,
-        __order_by => "sys_created_on");
-    my @recs = $qry->fetchAll();
-
-The following example retrieves all keys for the C<cmdb_ci>
-table in chunks of 100,000 at a time.
-
-    my $sn = ServiceNow($instance, $user, $pass, query => 100000);
-    my $cmdb_ci = $sn->table("cmdb_ci");
-    my $qry = $cmdb_ci->query();
-
-=cut
-
-sub query {
-    my $table = shift;
-    my $session = $table->{session};
-    my $limit = $session->{query};
-    my $query = ServiceNow::SOAP::Query->new($table);
-    # if there is a single parameter it must be an encoded query
-    if (@_ == 1) { unshift @_, '__encoded_query' };
-    my %params = @_;
-    my @keys;
-    if ($limit) {
-        my $first = 0;
-        my $last = $limit;
-        while ($first <= @keys) {
-            my %p = %params;
-            $p{__first_row} = $first;
-            $p{__last_row} = $last;
-            my @k = $table->getKeys(%p);
-            push @keys, @k;
-            $first += $limit;
-            $last = $first + $limit;
-        }
-    }
-    else {
-        @keys = $table->getKeys(%params);
-    }
-    # parameters which affect order or columns must be preserved
-    # so that they can be passed to getRecords
-    for my $k (keys %params) {
-        delete $params{$k} unless $k =~ /^(__order|__exclude|__use_view)/;
-    }
-    $query->{keys} = \@keys;
-    $query->{count} = scalar(@keys);
-    $query->{params} = \%params;
-    $query->{index} = 0;
-    return $query;
+sub getName {
+    return shift->{name};
 }
 
 =head2 getVariables
@@ -1260,6 +1197,10 @@ This method returns a list of hashes
 of the variables attached to a Requested Item (RITM).
 
 B<Note>: This function currently works only for the C<sc_req_item> table.
+This function requires read access to the three tables 
+C<item_option_new>,
+C<sc_item_option> and
+C<sc_item_option_mtom>.
 
 Each hash contains the following four fields:
 
@@ -1300,7 +1241,9 @@ B<Example>
     my $ritmRec = $sc_req_item->getRecord(number => $ritm_number);
     my @vars = $sc_req_item->getVariables($ritmRec->{sys_id});
     foreach my $var (sort { $a->{order} <=> $b->{order} } @vars) {
-        print "$var->{name} = $var->{value}\n";
+        my $name = $var->{name};
+        my $value = $var->{value};
+        print "$name = $value\n";
     }
 
 =cut
@@ -1311,12 +1254,9 @@ sub getVariables {
     my $sysid = shift;
     Carp::croak "getVariables: invalid sys_id: $sysid" unless _isGUID($sysid);
     my $session = $self->{session};
-    my $sc_item_option = $self->{sc_item_option} ||
-        ($self->{sc_item_option} = $session->table('sc_item_option'));
-    my $sc_item_option_mtom = $self->{sc_item_option_mtom} ||
-        ($self->{sc_item_option_mtom} = $session->table('sc_item_option_mtom'));
-    my $item_option_new = $self->{item_option_new} ||
-        ($self->{item_option_new} = $session->table('item_option_new'));
+    my $sc_item_option      = $session->private('sc_item_option');
+    my $sc_item_option_mtom = $session->private('sc_item_option_mtom');
+    my $item_option_new     = $session->private('item_option_new');
     my @vmtom = $sc_item_option_mtom->getRecords('request_item', $sysid);
     my $vrefkeys = join ',', map { $_->{sc_item_option} } @vmtom;
     my @vref = $sc_item_option->getRecords('sys_idIN' . $vrefkeys);
@@ -1339,6 +1279,219 @@ sub getVariables {
     return @result;
 }
 
+=head2 getWorkNotes
+
+This method retrieves all work notes for a task.
+It returns an array of records, each of which contains a C<value> field.
+
+See also: L</addWorkNote>, L</getJournalEntries>
+
+B<Example>
+
+    my @notes = $incident_tbl->getWorkNotes($sys_id);
+    foreach my $note (@notes) {
+        print $note->{value}, "\n";
+    }
+    
+=cut
+
+sub getWorkNotes {
+    my ($self, $sysid) = @_;
+    return $self->getJournalEntries($sysid, 'work_notes');
+}
+
+=head2 insert
+
+This method inserts a record.
+In a scalar context, this method returns a sys_id only.
+In a list context, this method returns a list of name/value pairs.
+
+For information on available parameters
+refer to the ServiceNow documentation on the 
+L<"insert" Direct SOAP API method|http://wiki.servicenow.com/index.php?title=SOAP_Direct_Web_Service_API#insert>.
+
+B<Syntax>
+
+    $sys_id = $table->insert(%values);
+    %result = $table->insert(%values);
+    
+B<Examples>
+
+    my $incident_tbl = $sn->table("incident");
+
+In a scalar context, the function returns a sys_id.
+
+    my $sys_id = $incident_tbl->insert(
+        short_description => $short_description,
+        assignment_group => "Network Support",
+        impact => 3);
+    print "sys_id=", $sys_id, "\n";
+
+In a list context, the function returns a list of name/value pairs.
+
+    my %result = $incident_tbl->insert(
+        short_description => $short_description,
+        assignment_group => "Network Support",
+        impact => 3);
+    print "number=", $result{number}, "\n";
+    print "sys_id=", $result{sys_id}, "\n";
+
+You can also call it this way:
+
+    my %rec;
+    $rec{short_description} = $short_description;
+    $rec{assignment_group} = "Network Support";
+    $rec{impact} = 3;
+    my $sys_id = $incident_tbl->insert(%rec);
+
+Note that for reference fields (e.g. assignment_group, assigned_to)
+you may pass in either a sys_id or a display value.
+
+=cut
+
+sub insert {
+    my $self = shift;
+    $self->traceBefore('insert');
+    my $som = $self->callMethod('insert', _params @_);
+    my $response = $som->body->{insertResponse};
+    my $sysid = $response->{sys_id};
+    $self->traceAfter();
+    return wantarray ? %{$response} : $sysid;
+}
+
+=head2 insertMultiple
+
+This method inserts multiple records.
+The input is an array of hash references.
+It returns an array of hash references.
+
+This method requires installation of the 
+L<ServiceNow Insert Multiple Web Services Plugin|http://wiki.servicenow.com/index.php?title=Web_Service_Import_Sets#Inserting_Multiple_Records>.
+
+For additional information 
+refer to the ServiceNow documentation on the 
+L<"insertMultiple" Direct SOAP API method|http://wiki.servicenow.com/index.php?title=SOAP_Direct_Web_Service_API#insertMultiple>.
+
+B<Syntax>
+
+    @results = $table->insertMultiple(@records);
+
+B<Example>
+
+    my $incident_tbl = $sn->table("incident");
+    my @allrecs;
+    foreach my $descr (@descriptions) {
+        my $newrec = {
+            short_description => $desc,
+            assignment_group => $groupName,
+            description => $desc,
+            impact => 3};
+        push @allrecs, $newrec;
+    }
+    my @results = $incident_tbl->insertMultiple(@allrecs);
+    foreach my $result (@results) {
+        print "inserted ", $result->{number}, "\n";
+    }
+
+=cut
+
+sub insertMultiple {
+    my $self = shift;
+    $self->traceBefore('insertMultiple');
+    my @params;
+    foreach my $rec (@_) {
+        Carp::croak "Expected array of hashs" unless ref $rec eq 'HASH';
+        push @params, SOAP::Data->name('record' => $rec);
+    }
+    my $som = $self->callMethod('insertMultiple', @params);
+    my $response = $som->body->{insertMultipleResponse};
+    my $insertResponse = $response->{insertResponse};
+    my @status = @{$insertResponse};
+    my $count = scalar(@status);
+    $self->traceAfter("$count records inserted");
+    return @status;
+}
+
+=head2 query
+
+This method creates a new L<Query|/ServiceNow::SOAP::Query> object
+by calling L</getKeys>.
+You can then use L</fetch> or L</fetchAll>
+to retrieve the actual record.
+
+This method normally fetches all keys using a single 
+call to L</getKeys>.
+However, if the number of keys is too large you may
+encounter timeout errors. In this situation the recommendation
+is to fetch the keys in chunks by specifying 
+a C<query> option in the L</ServiceNow> function.
+The suggested initial setting is C<query =E<gt> 100000>.
+If timeouts are still a problem then reduce the value further.
+
+B<Syntax>
+
+    $query = $table->query(%parameters);
+    $query = $table->query($encodedquery);
+    
+B<Example>
+
+The following example builds a list of all Incidents 
+created between 1/1/2014 and 2/1/2014 sorted by creation date/time.
+The L</fetchAll> method fetches chunks of records 
+until all records have been retrieved.
+
+    my $filter = "sys_created_on>=2014-01-01^sys_created_on<2014-02-01";
+    my $qry = $sn->table("incident")->query(
+        __encoded_query => $filter,
+        __order_by => "sys_created_on");
+    my @recs = $qry->fetchAll();
+
+The following example retrieves all keys for the C<cmdb_ci>
+table in chunks of 100,000 at a time.
+
+    my $sn = ServiceNow($instance, $user, $pass, query => 100000);
+    my $cmdb_ci_tbl = $sn->table("cmdb_ci");
+    my $qry = $cmdb_ci_tbl->query();
+
+=cut
+
+sub query {
+    my $table = shift;
+    my $session = $table->{session};
+    my $limit = $session->{query};
+    my $query = ServiceNow::SOAP::Query->new($table);
+    # if there is a single parameter it must be an encoded query
+    if (@_ == 1) { unshift @_, '__encoded_query' };
+    my %params = @_;
+    my @keys;
+    if ($limit) {
+        my $first = 0;
+        my $last = $limit;
+        while ($first <= @keys) {
+            my %p = %params;
+            $p{__first_row} = $first;
+            $p{__last_row} = $last;
+            my @k = $table->getKeys(%p);
+            push @keys, @k;
+            $first += $limit;
+            $last = $first + $limit;
+        }
+    }
+    else {
+        @keys = $table->getKeys(%params);
+    }
+    # parameters which affect order or columns must be preserved
+    # so that they can be passed to getRecords
+    for my $k (keys %params) {
+        delete $params{$k} unless $k =~ /^(__order|__exclude|__use_view)/;
+    }
+    $query->{keys} = \@keys;
+    $query->{count} = scalar(@keys);
+    $query->{params} = \%params;
+    $query->{index} = 0;
+    return $query;
+}
+
 =head2 setDV
 
 Used to enable (or disable) display values in queries.
@@ -1360,25 +1513,41 @@ B<Syntax>
 
 B<Examples>
 
+C<setDV("true")> causes display values to be returned
+for all fields instead of original values.
+
     my $sys_user_group = $sn->table("sys_user_group")->setDV("true");
     my $grp = $sys_user_group->getRecord(name => "Network Support");
     print "manager=", $grp->{manager}, "\n";
 
+C<setDV("all")> causes display values to be returned
+in addition to original values. 
+Display value fields have a "dv_" prefix.
+
     my $sys_user_group = $sn->table("sys_user_group")->setDV("all");
     my $grp = $sys_user_group->getRecord(name => "Network Support");
     print "manager=", $grp->{dv_manager}, "\n";
+
+Note that L</setDV> is a Table method, B<not> a Query method.
+
+    @recs = $table->setDV(1)->query()->fetchAll(); # Correct
+    @recs = $table->query()->setDV(1)->fetchAll(); # INCORRECT!
     
 =cut
 
 sub setDV {
     my ($self, $dv) = @_;
-    $dv = ''     if $dv eq '0';
-    $dv = 'true' if $dv eq '1';
-    $dv = 'all'  if $dv eq '2';
+    if ($dv eq '0') { $dv = '' }
+    elsif ($dv eq '1') { $dv = 'true' }
+    elsif ($dv eq '2') { $dv = 'all' };
     $self->{dv} = $dv;
     return $self;
 }
 
+sub getSession {
+    return $_[0]->{session};
+}
+
 sub setTrace {
     my ($self, $trace) = @_;
     $self->{trace} = $trace;
@@ -1417,6 +1586,58 @@ sub getFields {
 }
 
     
+=head2 update
+
+This method updates a record.
+For information on available parameters
+refer to the ServiceNow documentation on the 
+L<"update" Direct SOAP API method|http://wiki.servicenow.com/index.php?title=SOAP_Direct_Web_Service_API#update>.
+
+B<Syntax>
+
+    $table->update($sys_id, %parameters);
+    $table->update(%parameters);
+
+Note: 
+If the first syntax is used, then the C<sys_id>
+must be the first parameter.
+If the second syntax is used, 
+then the C<sys_id> must be included among the parameters.
+
+For reference fields (I<e.g.> assignment_group, assigned_to)
+you may pass in either a sys_id or a display value.
+
+B<Examples>
+
+The following three examples are equivalent.
+
+    $incident_tbl->update(
+        sys_id => $sys_id, 
+        assigned_to => "5137153cc611227c000bbd1bd8cd2005",
+        incient_state => 2);
+        
+    $incident_tbl->update(
+        sys_id => $sys_id, 
+        assigned_to => "Fred Luddy", 
+        incident_state => 2);
+    
+    $incident_tbl->update($sys_id, 
+        assigned_to => "Fred Luddy", 
+        incident_state => 2);
+
+=cut
+
+sub update {
+    my $self = shift;
+    if (_isGUID($_[0])) { unshift @_, 'sys_id' };
+    my %values = @_;
+    my $sysid = $values{sys_id};
+    $self->traceBefore("update");
+    my $som = $self->callMethod('update', _params @_);
+    $self->traceAfter();
+    return $self;
+}
+
 package ServiceNow::SOAP::Query;
 
 =head1 ServiceNow::SOAP::Query
@@ -1450,8 +1671,8 @@ the C<cmdb_rel_ci> table.
 The example involves two tables:
 C<cmdb_ci> and C<cmdb_rel_ci>.
 
-    my $cmdb_ci = $sn->table("cmdb_ci");
-    my $cmdb_rel_ci = $sn->table("cmdb_rel_ci");
+    my $cmdb_ci_tbl = $sn->table("cmdb_ci");
+    my $cmdb_rel_ci_tbl = $sn->table("cmdb_rel_ci");
     
 We assume that C<$parentKey> is the sys_id of a known configuration item.
 The objective is to print a list of all other configuration items 
@@ -1464,7 +1685,7 @@ C<@relData> is this list of C<cmdb_rel_ci> records.
 If there are more than 250 records, C<fetchAll> will loop internally 
 until all records have been retrieved.
 
-    my @relData = $cmdb_rel_ci->query(parent => $parentKey)->fetchAll();
+    my @relData = $cmdb_rel_ci_tbl->query(parent => $parentKey)->fetchAll();
 
 Next we extract the C<child> field from each of these C<cmdb_rel_ci>
 records to create a new list of sys_ids.
@@ -1475,7 +1696,7 @@ C<@childKeys> now contains a lists of sys_ids for configuration items.
 We convert this list of keys into a query object
 and fetch the records from the C<cmdb_ci> table.
     
-    @childRecs = $cmdb_ci->asQuery(@childKeys)->fetchAll();
+    @childRecs = $cmdb_ci_tbl->asQuery(@childKeys)->fetchAll();
     foreach my $childRec (@childRecs) {
         print $childRec->{name}, "\n";            
     }
@@ -1506,6 +1727,8 @@ or a string containing a comma delimited list of names.
 It will override any prior calls to L</exclude> or L</include>.
 It returns a reference to the modified Query object.
 
+See also: L</except>, L</include>
+
 B<Syntax>
 
     $query->exclude(@list_of_columns);
@@ -1642,6 +1865,7 @@ sub getKeys {
 
 This method will affect any subsequent calls to L</fetch>
 by limiting which columns are returned.
+This method utilizes the L</except> method documented above.
 By including only the needed columns in the query result,
 it is possible to significantly improve the performance of large queries.
 
@@ -1652,12 +1876,16 @@ It returns a reference to the modified Query object.
 
 For some reason the Direct Web Services API allows you to specify a 
 list of columns to be excluded from the query result,
-but there is no similar out-of-box capability to specify only the columns to be included.
-This function implements the more obviously needed behavior by inverting the list.
+but there is no similar out-of-box capability 
+to specify only the columns to be included.
+This function implements the more obviously needed behavior 
+by inverting the list.
 It uses the WSDL to generate a list of columns returned by L</getRecords>,
 and subtracts the specified names to create an C<"__exclude_columns">
 extended query parameter.
 
+See also: L</except>, L</exclude>
+
 B<Syntax>
 
     $query->include(@list_of_columns);
@@ -1697,7 +1925,7 @@ Sometimes, when you are developing a new script,
 it seems to hang at a certain point, 
 and you just want to know what it is doing.
 You can enable tracing of Web Service calls
-by setting the C<trace> parameter
+by setting the C<trace> option
 in the L</ServiceNow> function as follows.
 
     my $sn = ServiceNow($instance, $username, $password, trace => 1);
@@ -1707,7 +1935,7 @@ Set trace to 2 to print the complete XML result for each call.
 
 If you want even more, then add the following to your code.
 This will cause L<SOAP::Lite> to dump the HTTP headers
-and contents for all messages, both sent and received.
+and content for all messages, both sent and received.
 
     SOAP::Lite->import(+trace => 'debug');
 
